#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "FrameL.h"
#include "FrSegListTool.h" 


/*-------------------------------------------------------------FrSegListAdd--*/
int FrSegListAdd(FrSegList *segList, double tStart, double tEnd)
/*---------------------------------------------------------------------------*/
{ 
  if(segList == NULL) return(-1);
  if(tStart > tEnd)   return(-2);
  FrVectFillD(segList->tStart, tStart);
  FrVectFillD(segList->tEnd,   tEnd);
  return(0);
}
/*-------------------------------------------------------FrSegListBuildVect--*/
FrVect* FrSegListBuildVect(FrSegList *segList, 
                           double tStart, 
                           double length,
		           int nData)
/*---------------------------------------------------------------------------*/
{FrVect *vect;
  double dt;
  int i;

  if(segList == NULL) return(NULL);
  if(nData <= 0) return(NULL);
  dt = length/nData;

  vect = FrVectNew1D("Segment",FR_VECT_8R,nData,dt,"time (s)","1 means OK");
  if(vect == NULL) return(NULL);

  vect->GTime = tStart;
  for(i=0; i<nData; i++)
    {vect->dataD[i]= FrSegListCoverage(segList, tStart + i*dt, dt);}

  return(vect);
}
/*--------------------------------------------------------FrSegListCoverage--*/
double FrSegListCoverage(FrSegList *segList, double gtime, double dt)
/*---------------------------------------------------------------------------*/
/* This function compute the intersection between the segment list and the
   time interval. It returns a number from 0 to 1 (1 = 100% coverage).       */
/*---------------------------------------------------------------------------*/
{int i, iS, iE;
  double totTime, tStart, tEnd, length, use;

 if(segList == NULL) return(-1);

 iS = FrSegListFind(segList, gtime);
 iE = FrSegListFind(segList, gtime+dt);

 if(iS == iE)
   {if(iS >= 0) return(1.);
    else        return(0.);}

 totTime = 0;

 if(iS < 0) iS = -iS-1;
 if(iE < 0) iE = -iE-2;
 iE = iE + 1;
 if(iE > segList->tEnd->nData) iE = segList->tEnd->nData;

 for(i=iS; i<iE; i++)
    {tStart = segList->tStart->dataD[i];
     tEnd = segList->tEnd->dataD[i];
     if(tStart < gtime) tStart = gtime;
     if(tEnd > gtime+dt) tEnd = gtime+dt;
     length = tEnd - tStart;
     totTime += length;}
  
 use = totTime/dt;

 return(use);}

/*------------------------------------------------------------FrSegListDump--*/
int FrSegListDump(FrSegList *segList, FILE *fp, int debugLvl)
/*---------------------------------------------------------------------------*/
{int i, nData;
  double length, min, max, totTime, start, end, use;

  if(segList == NULL) return(-1);

  min = 1.e37;
  max = 0;
  totTime = 0;

  nData = segList->tStart->nData;
  for(i=0; i<nData; i++)
    {length = segList->tEnd->dataD[i]-segList->tStart->dataD[i];
      totTime += length;
     if(length > max) max = length;
     if(length < min) min = length;}
  
  fprintf(fp,"%8d segments;", nData);
  if(nData > 0) 
     {start = segList->tStart->dataD[0];
      end   = segList->tEnd->dataD[nData-1];
      fprintf(fp," start:%9.0f end:%9.0f",start, end);
      use = 100.*totTime/(end-start);}
  else {use = 0;}
  fprintf(fp," total time:%.3g (%.3g%%) for %s\n", totTime, use, segList->name);

  return(0);
}
/*------------------------------------------------------------FrSegListFind--*/
int FrSegListFind(FrSegList *segList,
                     double gtime)
/*---------------------------------------------------------------------------*/
/* This function search the segement index. 
                                    It returns a negative value is not found */
/*---------------------------------------------------------------------------*/
{int i, delta, nSegments;
 double *tStart, *tEnd;

 if(segList == NULL) return(-1);

 tStart = segList->tStart->dataD;
 tEnd   = segList->tEnd->dataD;
 nSegments = segList->tStart->nData;

 if(nSegments == 0)               return(-1);
 if(gtime < tStart[0])            return(-1);
 if(gtime >=   tEnd[nSegments-1]) return(-nSegments-1);
 if(gtime >= tStart[nSegments-1]) return(nSegments-1);

 delta = nSegments;
 i = delta/2;

 while(delta > 0)
   {if(i < 0) i = 0;
    if(i > nSegments-2) i = nSegments - 2;
    delta = delta-delta/2;
    
    if     (gtime < tStart[i]) {i = i - delta;}
    else if(gtime < tEnd[i])     {return(i);}
    else if(gtime < tStart[i+1]) {return(-i-2);}
    else {i = i + delta;}}

 return(-3);}

/*-------------------------------------------------------FrSegListFindFirst--*/
int FrSegListFindFirst(FrSegList *segList, double gtime, double dt)
/*---------------------------------------------------------------------------*/
/* This function check if there is a match between the segment list and the 
   time interval. It returns the id of the first segement or a negative value*/
/*---------------------------------------------------------------------------*/
{int i, j;

 if(segList == NULL) return(-1);
 
 i = FrSegListFind(segList, gtime);
 if(i>= 0) return(i);

 j = FrSegListFind(segList, gtime+dt);
 if(i == j) return(-2);
 
 return(-i-1);}


/*------------------------------------------------------------FrSegListFree--*/
void FrSegListFree(FrSegList *segList)
/*---------------------------------------------------------------------------*/
{
 if(segList == NULL) return;
 if(segList->next != NULL) FrSegListFree(segList->next);

 FrVectFree(segList->tStart);
 FrVectFree(segList->tEnd);
 if(segList->name != NULL) free(segList->name);
 free(segList);

 return;}
/*-------------------------------------------------------FrSegListIntersect--*/
FrSegList* FrSegListIntersect(FrSegList *list1, FrSegList* list2)
/*---------------------------------------------------------------------------*/
{FrSegList *list;
  int i,j;
  double *tStart1, *tStart2, *tEnd1, *tEnd2;

  if(list1 == NULL) return(NULL);
  if(list2 == NULL) return(NULL);

  list = FrSegListNew("intersect");
     
  tStart1 = list1->tStart->dataD;
  tStart2 = list2->tStart->dataD;
  tEnd1  = list1->tEnd->dataD;
  tEnd2  = list2->tEnd->dataD;

  i = 0;
  j = 0;
  while(i<list1->tStart->nData &&
        j<list2->tStart->nData)
    {if(tEnd1[i] < tStart2[j])        /*---case: seg 1 enf before seg 2 start*/
       {i++;}                         /*---case: 1111 22222 */
      else if(tEnd1[i] < tEnd2[j])    /*---case: seg 1 end before seg 2 end */
	{if(tStart1[i] < tStart2[j])  /*---case: 1111***222 */
	    {FrSegListAdd(list, tStart2[j], tEnd1[i]);
	      i++;}
	  else                        /*---case: 222****222 */
	    {FrSegListAdd(list, tStart1[i], tEnd1[i]);
	      i++;}}
      else                            /*---case: seg 1 end after seg 2*/
	{if(tStart1[i] < tStart2[j])    /*---case: 1111*****1111 */
	    {FrSegListAdd(list,tStart2[j],tEnd2[j]);
	      i++;}
	  else if(tStart1[i] < tEnd2[j]) /*---case: 222*****1111 */
	    {FrSegListAdd(list,tStart1[i],tEnd2[j]);
	      i++;}
	  else {j++;}}}                 /*---case: 2222  1111*/

  return(list);
}
/*-------------------------------------------------------------FrSegListNew--*/
FrSegList *FrSegListNew(char *name)
/*---------------------------------------------------------------------------*/
{FrSegList *segList;
 
 segList = (FrSegList *) calloc(1,sizeof(FrSegList));
 if(segList == NULL)   return(NULL);
 
 FrStrCpy(&(segList->name), name);

 segList->tStart = FrVectNew1D("start time",FR_VECT_8R,0,1.,NULL,"GPS time");
 segList->tEnd   = FrVectNew1D("start time",FR_VECT_8R,0,1.,NULL,"GPS time");
 if(segList->tEnd == NULL) return(NULL);

 return(segList);}

/*-----------------------------------------------------------FrSegListPrint--*/
int FrSegListPrint(FrSegList *segList, char *fileName)
/*---------------------------------------------------------------------------*/
{FILE *fp;
  int i, dec, nDigits;
  double *tStart, *tEnd;

  if(segList == NULL) return(-1);
  if(fileName == NULL) return(-1);

  /*---- get the needed resolution-----*/

  tStart = segList->tStart->dataD;
  tEnd   = segList->tEnd->dataD;

  nDigits = 0;
  for(i=0; i<segList->tStart->nData; i++)
    {dec = (int) 1.e6*(tStart[i]-((int)tStart[i]));
     if     (dec%1000      != 0) nDigits = 6;
     else if(dec%1000000   != 0 && nDigits < 3) nDigits = 3;
     
     dec = (int) 1.e6*(tEnd  [i]-((int)tEnd[i]));
     if     (dec%1000      != 0) nDigits = 6;
     else if(dec%1000000   != 0 && nDigits < 3) nDigits = 3;}

  fp = fopen(fileName, "w");
  if(nDigits == 0)
      fprintf(fp,"    id       Start         End      length\n");
  else if(nDigits == 3)
      fprintf(fp,"    id       Start            End         length\n");
  else 
      fprintf(fp,"    id       Start               End            length\n");

  for(i=0; i<segList->tStart->nData; i++)
    {if(nDigits == 6) fprintf(fp, "%6d %17.6f %17.6f %17.6f\n", 
			      i, tStart[i], tEnd[i], tEnd[i]-tStart[i]);
     else if(nDigits == 3)  fprintf(fp, "%6d %14.3f %14.3f %14.3f\n", 
			      i, tStart[i], tEnd[i], tEnd[i]-tStart[i]);
     else                   fprintf(fp, "%6d %11.0f %11.0f %11.0f\n",
			      i, tStart[i], tEnd[i], tEnd[i]-tStart[i]);}
  
  fclose(fp);

  return(0);
}
/*------------------------------------------------------------FrSegListRead--*/
FrSegList* FrSegListRead(char *fileName)
/*---------------------------------------------------------------------------*/
{FILE *fp;
  FrSegList *segList;
  double tStart, t;
  char word[40];
  int i;

  if(fileName == NULL) return(NULL);

  segList = FrSegListNew(fileName);
  if(segList == NULL) return(NULL);

  tStart = 0;
  fp = fopen(fileName, "r");
  while(fscanf(fp,"%39s",word) == 1)
    {i = sscanf(word,"%lg",&t);
      if(i != 1) continue;
      if(t > 500123123)
	{if(tStart == 0) tStart = t;
	  else 
	    {FrVectFillD(segList->tStart, tStart);
	      FrVectFillD(segList->tEnd, t);
	      tStart = 0;}}}
  
  fclose(fp);

  return(segList);
}
/*----------------------------------------------------------FrSegListShrink--*/
int FrSegListShrink(FrSegList *segList, double tStart, double tEnd)
/*---------------------------------------------------------------------------*/
{FrSegList *limit, *intersect;

  if(segList == NULL) return(-1);

  limit = FrSegListNew("temporary");
  if(limit == NULL) return(-2);
  FrSegListAdd(limit, tStart, tEnd);
  intersect = FrSegListIntersect(segList, limit);
  FrVectFree(segList->tStart);
  FrVectFree(segList->tEnd);

  segList->tStart = intersect->tStart;
  segList->tEnd   = intersect->tEnd;
  free(intersect->name);
  free(intersect);

  FrSegListFree(limit);

  return(segList->tStart->nData);
}
